import json
import bisect


class misc:
    def chunk(iterator, count):
        itr = iter(iterator)
        while True:
            try:
                yield tuple([next(itr) for i in range(count)])
            except StopIteration:
                return

    def seek_patch_data(patches, offset, bytes):
        """[summary]

        Arguments:
            patches {list} -- a list of dictionaries depicting raw patch data
            offset {int} -- a decimal integer of the offset to look for
            bytes {int} -- the number of bytes to retrieve

        Raises:
            ValueError -- raised if the offset could not be found

        Returns:
            list -- a list of bytes of the requested offset
        """

        offsetlist = []
        for patch in patches:
            for key, value in patch.items():
                offsetlist.append(int(key))
        offsetlist_sorted = sorted(offsetlist)
        i = bisect.bisect_left(offsetlist_sorted, offset)
        if i:
            if offsetlist_sorted[i] == offset:
                seek = str(offset)
                for patch in patches:
                    if seek in patch:
                        return patch[seek][:bytes]
            else:
                left_slice = offset - offsetlist_sorted[i - 1]
                for patch in patches:
                    seek = str(offsetlist_sorted[i - 1])
                    if seek in patch:
                        return patch[seek][left_slice:left_slice + bytes]
        raise ValueError



from .misc import misc
from .exceptions import alttprException

import hashlib
import aiofiles


class romfile:
    async def read(srcfilepath, verify_checksum=True):
        """Reads a ROM and converts it to a list of bytes, preparing it for patching.

        Arguments:
            srcfilepath {str} -- A path (either relative or absolute) to a file on the filesystem to be read.
            verify_checksum {bool} -- Verify the checksum of the file to ensure it is an unheadered ALTTPR Japan 1.0 ROM.

        Raises:
            alttprException -- Returns if the sha256 checksum does not match a ALTTPR Japan 1.0 ROM.

        Returns:
            list -- A list of bytes depicting the read ROM file.
        """
        async with aiofiles.open(srcfilepath, "rb") as f:
            baserom_array = list(await f.read())
        if verify_checksum:
            if len(baserom_array) == 1049088:
                baserom_array = baserom_array[512:]
            expected_rom_sha256 = '794e040b02c7591b59ad8843b51e7c619b88f87cddc6083a8e7a4027b96a2271'
            sha256_hash = hashlib.sha256()
            sha256_hash.update(bytes(baserom_array))
            if not sha256_hash.hexdigest() == expected_rom_sha256:
                raise alttprException(
                    'Expected checksum "{expected_rom_sha256}", got "{actual_checksum}" instead.  Verify the source ROM is an unheadered Japan 1.0 Link to the Past ROM.'.format(
                        expected_rom_sha256=expected_rom_sha256, actual_checksum=sha256_hash.hexdigest()))
        return baserom_array

    async def write(rom, dstfilepath):
        fw = await aiofiles.open(dstfilepath, "wb")
        patchrom = bytes()
        for idx, chunk_array in enumerate(misc.chunk(rom, 256)):
            patchrom += bytes(chunk_array)
        await fw.write(patchrom)
        fw.close



import itertools
from .exceptions import alttprException


class patch:
    def apply(rom, patches):
        """Applies a patch, which is a list of dictionaries

        Arguments:
            rom {list} -- A list of bytes depicting the ROM data to be patched.
            patches {list} -- A list of dictionaries that depict of set of patches to be applied to the ROM.

        Returns:
            list -- a list of bytes depicitng the patched rom
        """
        for patch in patches:
            offset = int(list(patch.keys())[0])
            patch_values = list(patch.values())[0]
            for idx, value in enumerate(patch_values):
                rom[offset + idx] = value
        return rom

    def heart_speed(speed='half'):
        """Set the low-health warning beep interval.

        Keyword Arguments:
            speed {str} -- Chose the speed at which the low health warning beeps.
                Options are 'off', 'double', 'normal', 'half', and 'quarter'. (default: {'half'})

        Returns:
            list -- a list of dictionaries indicating which ROM address offsets to write and what to write to them
        """
        if speed is None:
            speed = 'normal'
        sbyte = {
            'off': 0,
            'double': 16,
            'normal': 32,
            'half': 64,
            'quarter': 128,
        }
        patch = [{
            '1572915': [sbyte[speed]]
        }]
        return patch

    def heart_color(color='red'):
        """Set the color of the hearts on the player's HUD.

        Keyword Arguments:
            color {str} -- The heart color.  Options are 'red', 'blue', 'green', and 'yellow' (default: {'red'})

        Returns:
            list -- a list of dictionaries indicating which ROM address offsets to write and what to write to them
        """

        if color is None:
            color = 'red'
        cbyte = {
            'blue': [44, 13],
            'green': [60, 25],
            'yellow': [40, 9],
            'red': [36, 5],
        }
        byte = cbyte[color][0]
        file_byte = cbyte[color][1]
        patch = [
            {'457246': [byte]},
            {'457248': [byte]},
            {'457250': [byte]},
            {'457252': [byte]},
            {'457254': [byte]},
            {'457256': [byte]},
            {'457258': [byte]},
            {'457260': [byte]},
            {'457262': [byte]},
            {'457264': [byte]},
            {'415073': [file_byte]},
        ]
        return patch

    def music(music=True):
        """Enables, or disables, the in-game music.  Useful if you want to use an MSU-1 soundtrack instead.

        Keyword Arguments:
            music {bool} -- If true, music is enabled.  If false, the music id disabled. (default: {True})

        Returns:
            list -- a list of dictionaries indicating which ROM address offsets to write and what to write to them
        """

        if music:
            return [{'1573402': [0]}]
        else:
            return [{'1573402': [1]}]

    def sprite(spr):
        """Creates a patch for to replace Link's sprite with the contents of a XSPR or SPR file.

        Arguments:
            spr {list} -- a list of bytes that depicts a ZSPR or SPR file

        Returns:
            list -- a list of dictionaries indicating which ROM address offsets to write and what to write to them
        """

        if spr[:4] == [90, 83, 80, 82]:
            # stolen from VT's code
            gfx_offset = spr[12] << 24 | spr[11] << 16 | spr[10] << 8 | spr[9]
            palette_offset = spr[18] << 24 | spr[17] << 16 | spr[16] << 8 | spr[15]
            patch = [
                {'524288': spr[gfx_offset:gfx_offset + 28671]},
                {'905992': spr[palette_offset:palette_offset + 120]},
                {'912885': spr[palette_offset + 120:palette_offset + 120 + 3]}
            ]
        # Else treat it like a SPR file instead
        else:
            patch = [
                {'524288': spr[0:28671]},
                {'905992': spr[28672:28791]},
                {
                    '912885': [
                        spr[28726],
                        spr[28727],
                        spr[28756],
                        spr[28757],
                    ]
                }
            ]
        return patch

    def checksum(rom):
        """Writes a patch that fixes a ROM's checksum.  This should be the last patch applied to a ROM before it is written.

        Arguments:
            rom {list} -- a list of bytes depicitng the rom

        Returns:
             list -- a list of dictionaries indicating which ROM address offsets to write and what to write to them
        """

        sum_of_bytes = sum(rom[:32731]) + sum(rom[32736:])
        checksum = (sum_of_bytes + 510) & 65535
        inverse = checksum ^ 65535
        patch = [
            {
                '32732': [
                    inverse & 255,
                    inverse >> 8,
                    checksum & 255,
                    checksum >> 8,
                ]
            }
        ]
        return patch

    def expand(rom, newlenmb):
        """Expands the byte list of a ROM to the specified number of megabytes, filling in the new space with zeroes.

        Arguments:
            rom {list} -- a list of bytes depicitng the rom

        Keyword Arguments:
            newlenmb {int} -- The size of the ROM should be, in megabytes.

        Raises:
            alttprException -- Raised if the new length is shorter than the current size of the byte list.

        Returns:
            list -- a list of bytes depicitng the rom
        """

        newlen = int(newlenmb) * 1024 * 1024
        if len(rom) > newlen:
            raise alttprException('ROM is already larger than {bytes}'.format(
                bytes=newlen
            ))
        diff = len(rom) - newlen
        if diff > 0:
            rom[newlen] = 0
        else:
            rom.extend(itertools.repeat(0, -diff))
            rom.append(0)
        return rom



import math


class customizer:
    def convert2settings(
            customizer_save,
            tournament=False,
            spoilers="off",
            spoilers_ongen=False):
        """Converts a customizer-settings.json file from alttpr.com to a settings dictionary that can be used for generating a game.

        Arguments:
            customizer_save {dict} -- A dictionary of the customizer-settings.json file (needs to already be converted to a dict)

        Keyword Arguments:
            tournament {bool} -- Setting to True generates a race rom, which excludes the spoiler log. (default: {False})
            spoilers {str} -- Sets the spoiler mode. (default: {"off"})
            spoielrs_ongen {bool} -- Sets spoiler mode to "generate".  This is deprecated. (default: {False})

        Returns:
            dict -- a dictionary of settings that can be used with pyz3r.alttpr()
        """

        if spoilers_ongen: spoilers="generate"

        # the settings defaults, hopefully this is accurate

        settings = {
            "glitches": "none",
            "item_placement": "advanced",
            "dungeon_items": "standard",
            "accessibility": "items",
            "goal": "ganon",
            "crystals": {
                "ganon": "7",
                "tower": "7"
            },
            "mode": "open",
            "weapons": "randomized",
            "item": {
                "pool": "normal",
                "functionality": "normal"
            },
            "tournament": tournament,
            "spoilers": spoilers,
            "lang": "en",
            "name": "",
            "notes": "",
            "l": {},
            "eq": ["BossHeartContainer", "BossHeartContainer", "BossHeartContainer"],
            "drops": {
                "0": ["auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill"],
                "1": ["auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill"],
                "2": ["auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill"],
                "3": ["auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill"],
                "4": ["auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill"],
                "5": ["auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill"],
                "6": ["auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill", "auto_fill"],
                "pull": ["auto_fill", "auto_fill", "auto_fill"],
                "crab": ["auto_fill", "auto_fill"],
                "stun": ["auto_fill"],
                "fish": ["auto_fill"]
            },
            "custom": {
                "item.Goal.Required": "",
                "item.require.Lamp": False,
                "item.value.BlueClock": "",
                "item.value.GreenClock": "",
                "item.value.RedClock": "",
                "item.value.Rupoor": "",
                "prize.crossWorld": True,
                "prize.shuffleCrystals": True,
                "prize.shufflePendants": True,
                "region.bossNormalLocation": True,
                "region.wildBigKeys": False,
                "region.wildCompasses": False,
                "region.wildKeys": False,
                "region.wildMaps": False,
                "rom.dungeonCount": "off",
                "rom.freeItemMenu": False,
                "rom.freeItemText": False,
                "rom.mapOnPickup": False,
                "rom.timerMode": "off",
                "rom.timerStart": "",
                "rom.rupeeBow": False,
                "rom.genericKeys": False,
                "spoil.BootsLocation": False,
                "spoil.Hints": True,
                "canBombJump": False,
                "canBootsClip": False,
                "canBunnyRevive": False,
                "canBunnySurf": False,
                "canDungeonRevive": False,
                "canFakeFlipper": False,
                "canMirrorClip": False,
                "canMirrorWrap": False,
                "canOneFrameClipOW": False,
                "canOWYBA": False,
                "canSuperBunny": False,
                "canSuperSpeed": False,
                "canWaterFairyRevive": False,
                "canWaterWalk": False,
                "item": {
                    "count": {
                        "BottleWithRandom": 4,
                        "Nothing": 0,
                        "L1Sword": 0,
                        "L1SwordAndShield": 0,
                        "MasterSword": 0,
                        "L3Sword": 0,
                        "L4Sword": 0,
                        "BlueShield": 0,
                        "RedShield": 0,
                        "MirrorShield": 0,
                        "FireRod": 1,
                        "IceRod": 1,
                        "Hammer": 1,
                        "Hookshot": 1,
                        "Bow": 0,
                        "Boomerang": 1,
                        "Powder": 1,
                        "Bombos": 1,
                        "Ether": 1,
                        "Quake": 1,
                        "Lamp": 1,
                        "Shovel": 1,
                        "OcarinaInactive": 1,
                        "CaneOfSomaria": 1,
                        "Bottle": 0,
                        "PieceOfHeart": 24,
                        "CaneOfByrna": 1,
                        "Cape": 1,
                        "MagicMirror": 1,
                        "PowerGlove": 0,
                        "TitansMitt": 0,
                        "BookOfMudora": 1,
                        "Flippers": 1,
                        "MoonPearl": 1,
                        "BugCatchingNet": 1,
                        "BlueMail": 0,
                        "RedMail": 0,
                        "Bomb": 0,
                        "ThreeBombs": 16,
                        "Mushroom": 1,
                        "RedBoomerang": 1,
                        "BottleWithRedPotion": 0,
                        "BottleWithGreenPotion": 0,
                        "BottleWithBluePotion": 0,
                        "TenBombs": 1,
                        "OneRupee": 2,
                        "FiveRupees": 4,
                        "TwentyRupees": 28,
                        "BowAndArrows": 0,
                        "BowAndSilverArrows": 0,
                        "BottleWithBee": 0,
                        "BottleWithFairy": 0,
                        "BossHeartContainer": 10,
                        "HeartContainer": 1,
                        "OneHundredRupees": 1,
                        "FiftyRupees": 7,
                        "Heart": 0,
                        "Arrow": 1,
                        "TenArrows": 12,
                        "SmallMagic": 0,
                        "ThreeHundredRupees": 5,
                        "BottleWithGoldBee": 0,
                        "OcarinaActive": 0,
                        "PegasusBoots": 1,
                        "BombUpgrade5": 0,
                        "BombUpgrade10": 0,
                        "ArrowUpgrade5": 0,
                        "ArrowUpgrade10": 0,
                        "HalfMagic": 1,
                        "QuarterMagic": 0,
                        "SilverArrowUpgrade": 0,
                        "Rupoor": 0,
                        "RedClock": 0,
                        "BlueClock": 0,
                        "GreenClock": 0,
                        "ProgressiveSword": 4,
                        "ProgressiveShield": 3,
                        "ProgressiveArmor": 2,
                        "ProgressiveGlove": 2,
                        "ProgressiveBow": 2,
                        "Triforce": 0,
                        "TriforcePiece": 0,
                        "MapA2": 1,
                        "MapD7": 1,
                        "MapD4": 1,
                        "MapP3": 1,
                        "MapD5": 1,
                        "MapD3": 1,
                        "MapD6": 1,
                        "MapD1": 1,
                        "MapD2": 1,
                        "MapA1": 0,
                        "MapP2": 1,
                        "MapP1": 1,
                        "MapH1": 0,
                        "MapH2": 1,
                        "CompassA2": 1,
                        "CompassD7": 1,
                        "CompassD4": 1,
                        "CompassP3": 1,
                        "CompassD5": 1,
                        "CompassD3": 1,
                        "CompassD6": 1,
                        "CompassD1": 1,
                        "CompassD2": 1,
                        "CompassA1": 0,
                        "CompassP2": 1,
                        "CompassP1": 1,
                        "CompassH1": 0,
                        "CompassH2": 0,
                        "BigKeyA2": 1,
                        "BigKeyD7": 1,
                        "BigKeyD4": 1,
                        "BigKeyP3": 1,
                        "BigKeyD5": 1,
                        "BigKeyD3": 1,
                        "BigKeyD6": 1,
                        "BigKeyD1": 1,
                        "BigKeyD2": 1,
                        "BigKeyA1": 0,
                        "BigKeyP2": 1,
                        "BigKeyP1": 1,
                        "BigKeyH1": 0,
                        "BigKeyH2": 0,
                        "KeyH2": 1,
                        "KeyH1": 0,
                        "KeyP1": 0,
                        "KeyP2": 1,
                        "KeyA1": 2,
                        "KeyD2": 1,
                        "KeyD1": 6,
                        "KeyD6": 3,
                        "KeyD3": 3,
                        "KeyD5": 2,
                        "KeyP3": 1,
                        "KeyD4": 1,
                        "KeyD7": 4,
                        "KeyA2": 4
                    }
                },
                "drop": {
                    "count": {
                        "Bee": 0,
                        "BeeGood": 0,
                        "Heart": 13,
                        "RupeeGreen": 9,
                        "RupeeBlue": 7,
                        "RupeeRed": 6,
                        "BombRefill1": 7,
                        "BombRefill4": 1,
                        "BombRefill8": 2,
                        "MagicRefillSmall": 6,
                        "MagicRefillFull": 3,
                        "ArrowRefill5": 5,
                        "ArrowRefill10": 3,
                        "Fairy": 1
                    }
                }
            }
        }

        try:
            if not customizer_save['randomizer.glitches_required'] is None:
                settings['glitches'] = customizer_save['randomizer.glitches_required']
        except KeyError:
            pass

        try:
            if not customizer_save['randomizer.accessibility'] is None:
                settings['accessibility'] = customizer_save['randomizer.accessibility']
        except KeyError:
            pass

        try:
            if not customizer_save['randomizer.goal'] is None:
                settings['goal'] = customizer_save['randomizer.goal']
        except KeyError:
            pass

        try:
            if not customizer_save['randomizer.tower_open'] is None:
                settings['crystals']['tower'] = customizer_save['randomizer.tower_open']
        except KeyError:
            pass

        try:
            if not customizer_save['randomizer.ganon_open'] is None:
                settings['crystals']['ganon'] = customizer_save['randomizer.ganon_open']
        except KeyError:
            pass

        try:
            if not customizer_save['randomizer.dungeon_items'] is None:
                settings['dungeon_items'] = customizer_save['randomizer.dungeon_items']
        except KeyError:
            pass

        try:
            if not customizer_save['randomizer.item_placement'] is None:
                settings['item_placement'] = customizer_save['randomizer.item_placement']
        except KeyError:
            pass

        try:
            if not customizer_save['randomizer.world_state'] is None:
                settings['mode'] = customizer_save['randomizer.world_state']
        except KeyError:
            pass

        try:
            if not customizer_save['randomizer.hints'] is None:
                settings['custom']['spoil.Hints'] = customizer_save['randomizer.hints']
        except KeyError:
            pass

        try:
            if not customizer_save['randomizer.weapons'] is None:
                settings['weapons'] = customizer_save['randomizer.weapons']
        except KeyError:
            pass

        try:
            if not customizer_save['randomizer.item_pool'] is None:
                settings['item']['pool'] = customizer_save['randomizer.item_pool']
        except KeyError:
            pass

        try:
            if not customizer_save['randomizer.item_functionality'] is None:
                settings['item']['functionality'] = customizer_save['randomizer.item_functionality']
        except KeyError:
            pass

        # vt.custom.prizepacks
        try:
            if not customizer_save['vt.custom.prizepacks'] is None:
                settings['drops'] = customizer_save['vt.custom.prizepacks']
        except KeyError:
            pass

        # vt.custom.drops
        try:
            if not customizer_save['vt.custom.drops'] is None:
                settings['drop']['count'] = customizer_save['vt.custom.drops']
        except KeyError:
            pass

        try:
            if not customizer_save['vt.custom.settings'] is None:
                for key, value in customizer_save['vt.custom.settings'].items(
                ):
                    settings['custom'][key] = value
        except KeyError:
            pass

        try:
            if not customizer_save['vt.custom.glitches'] is None:
                for key, value in customizer_save['vt.custom.glitches'].items(
                ):
                    settings['custom'][key] = value
        except KeyError:
            pass


        try:
            if not customizer_save['vt.custom.items'] is None:
                settings['custom']['item']['count'] = customizer_save['vt.custom.items']
        except KeyError:
            pass

        try:
            if not customizer_save['vt.custom.locations'] is None:
                settings['l'] = customizer_save['vt.custom.locations']
        except KeyError:
            pass

        try:
            if not customizer_save['vt.custom.notes'] is None:
                settings['notes'] = customizer_save['vt.custom.notes']
        except KeyError:
            pass

        try:
            if not customizer_save['vt.custom.name'] is None:
                settings['name'] = customizer_save['vt.custom.name']
        except KeyError:
            pass

        try:
            if not customizer_save['vt.custom.equipment'] is None:
                eq = []
                for key, value in customizer_save['vt.custom.equipment'].items(
                ):
                    if key in ['Bottle1', 'Bottle2', 'Bottle3', 'Bottle4']:
                        if value == 1:
                            eq += ['Bottle']
                        elif value == 2:
                            eq += ['BottleWithRedPotion']
                        elif value == 3:
                            eq += ['BottleWithGreenPotion']
                        elif value == 4:
                            eq += ['BottleWithBluePotion']
                        elif value == 5:
                            eq += ['BottleWithBee']
                        elif value == 6:
                            eq += ['BottleWithGoldBee']
                        elif value == 7:
                            eq += ['BottleWithFairy']
                    elif key == 'ProgressiveBow':
                        if value == 1:
                            eq += ['SilverArrowUpgrade']
                        elif value == 2:
                            eq += ['Bow']
                        elif value == 3:
                            eq += ['BowAndSilverArrows']
                    elif key == 'Boomerang':
                        if value == 1:
                            eq += ['Boomerang']
                        elif value == 2:
                            eq += ['RedBoomerang']
                        elif value == 3:
                            eq += ['Boomerang', 'RedBoomerang']
                    elif key == 'Ocarina':
                        if value == 1:
                            eq += ['OcarinaInactive']
                        elif value == 2:
                            eq += ['OcarinaActive']
                    elif key == "Rupees":
                        value = int(
                            customizer_save['vt.custom.equipment']['Rupees'])

                        eq += math.floor(value / 300) * ['ThreeHundredRupees']
                        value %= 300

                        eq += math.floor(value / 100) * ['OneHundredRupees']
                        value %= 100

                        eq += math.floor(value / 50) * ['FiftyRupees']
                        value %= 50

                        eq += math.floor(value / 20) * ['TwentyRupees']
                        value %= 20

                        eq += math.floor(value / 5) * ['FiveRupees']
                        value %= 5

                        eq += math.floor(value / 1) * ['OneRupee']
                    else:
                        eq += int(value) * [key]

                settings['eq'] = eq

        except KeyError:
            pass

        return settings
